# Copyright 2017 Kakao
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
import os
import glob
import platform
import subprocess

from Cython.Build import cythonize
from setuptools import Extension, setup

NAME = 'n2'
VERSION = '0.1.6'


def long_description():
    with io.open('README.rst', 'r', encoding='utf-8') as f:
        readme = f.read()
    return readme


def set_binary_mac():
    gcc_dir = subprocess.check_output("brew --prefix gcc", shell=True).decode().strip()
    gcc_dir = os.path.join(gcc_dir, 'bin')
    gpp_binaries = glob.glob(os.path.join(gcc_dir, 'g++-[0-9]'))
    gcc_binaries = glob.glob(os.path.join(gcc_dir, 'gcc-[0-9]'))
    binaries = [gcc_binaries, gpp_binaries]
    targets = ["CC", "CXX"]
    fail = False
    for binary, target in zip(binaries, targets):
        if binary:
            binary = sorted(binary, key=lambda x: int(x.split('-')[1]))[-1]
            os.environ[target] = os.path.join(gcc_dir, binary)
        else:
            fail = True
            break

    if fail:
        raise AttributeError('No GCC available. Install gcc from Homebrew using brew install gcc.')


def define_extensions(**kwargs):
    system = platform.system().lower()
    if "windows" in system:  # Windows
        raise AttributeError("Installation on Windows is not supported yet.")
    elif "darwin" in system:  # osx
        set_binary_mac()

    libraries = []
    extra_link_args = []
    extra_compile_args = ['-std=c++14', '-O3', '-fPIC', '-march=native']
    extra_link_args.append('-fopenmp')
    extra_compile_args.append('-fopenmp')

    sources = ['./src/heuristic.cc', './src/hnsw.cc', './src/hnsw_node.cc',
               './src/hnsw_build.cc', './src/hnsw_model.cc', './src/hnsw_search.cc',
               './src/mmap.cc', './bindings/python/n2.pyx']

    boost_dirs = ['assert', 'bind', 'concept_check', 'config', 'core', 'detail', 'heap', 'iterator', 'mp11', 'mpl',
                  'parameter', 'preprocessor', 'static_assert', 'throw_exception', 'type_traits', 'utility']
    include_dirs = ['./include/', './third_party/spdlog/include/', './third_party/eigen']
    include_dirs.extend(['third_party/boost/' + b + '/include/' for b in boost_dirs])

    client_ext = Extension(name='n2',
                           sources=sources,
                           extra_compile_args=extra_compile_args,
                           libraries=libraries,
                           extra_link_args=extra_link_args,
                           include_dirs=include_dirs,
                           language="c++",)
    return cythonize(client_ext)


setup(
    name=NAME,
    version=VERSION,
    description='Approximate Nearest Neighbor library',
    long_description=long_description(),
    author='Kakao.corp',
    author_email='recotech.kakao@gmail.com',
    license='Apache License 2.0',
    install_requires=[
        "cython",
    ],
    classifiers=[
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'Programming Language :: Cython',
        'Topic :: Software Development :: Libraries :: Python Modules'],

    keywords='Approximate Nearest Neighbor',
    ext_modules=define_extensions(),
)



import sys

import h5py
import numpy
from sklearn.model_selection import train_test_split


class BruteForceBLAS():
    def __init__(self, metric, precision=numpy.float32):
        self._metric = metric
        self._precision = precision

    def fit(self, X):
        lens = (X ** 2).sum(-1)
        if self._metric == 'angular':
            X /= numpy.sqrt(lens)[..., numpy.newaxis]
            self.index = numpy.ascontiguousarray(X, dtype=self._precision)
        elif self._metric == 'euclidean':
            self.index = numpy.ascontiguousarray(X, dtype=self._precision)
            self.lengths = numpy.ascontiguousarray(lens, dtype=self._precision)

    def query(self, v, n):
        v = numpy.ascontiguousarray(v, dtype=self._precision)
        if self._metric == 'angular':
            dists = -numpy.dot(self.index, v)
        elif self._metric == 'euclidean':
            dists = self.lengths - 2 * numpy.dot(self.index, v)
        indices = numpy.argpartition(dists, n)[:n]
        return sorted(indices, key=lambda index: dists[index])


def write_output(train, test, fn, distance, point_type='float', count=100):
    f = h5py.File(fn, 'w')
    f.attrs['distance'] = distance
    f.attrs['point_type'] = point_type
    f.create_dataset('train', (len(train), len(train[0])), dtype=train.dtype)[:] = train
    f.create_dataset('test', (len(test), len(test[0])), dtype=test.dtype)[:] = test
    neighbors = f.create_dataset('neighbors', (len(test), count), dtype='i')
    bf = BruteForceBLAS(distance, precision=train.dtype)
    bf.fit(train)
    for i, x in enumerate(test):
        sys.stderr.write('computing queries %d/%d ...\r' % (i+1, len(test)))
        neighbors[i] = bf.query(x, count)

    sys.stderr.write('\n')
    f.close()


def youtube(out_fn):
    fn = './data/youtube.txt'
    X = []
    for line in open(fn):
        v = [float(x) for x in line.strip().split()]
        X.append(numpy.array(v))

    X_train, X_test = train_test_split(X, test_size=10000, random_state=1)
    write_output(numpy.array(X_train), numpy.array(X_test), out_fn, 'angular')



if __name__ == '__main__':
    out_fn = os.path.join('data', 'youtube-40-angular.hdf5')
    youtube(out_fn)



import os
import math
import argparse
import matplotlib.pyplot as plt
from collections import defaultdict


def parse(fname):
    data = defaultdict(list)
    for line in open(fname):
        library, _, build_time, search_elapsed, accuracy, index_size_kb = line.strip().split('\t')
        data[library].append((float(accuracy), math.log(float(search_elapsed)), float(build_time), float(index_size_kb)))
    return data


def visualize(data, fname):
    fig = plt.figure(figsize=(8, 12))

    ax1 = fig.add_subplot(4, 1, 1)
    draw_build_time(ax1, data)

    ax2 = fig.add_subplot(4, 1, 2)
    draw_index_size(ax2, data)

    ax3 = fig.add_subplot(4, 1, 3)
    draw_accuracy_elapsed(ax3, data)

    ax4 = fig.add_subplot(4, 1, 4)
    draw_accuracy_elapsed(ax4, data, limit=0.8)

    fig.tight_layout()
    fig.savefig(fname)


def draw_build_time(ax, data):
    ax.set_title('build time')
    keys = data.keys()
    y = [max([t for _, _, t, _ in data[k]]) for k in keys]
    x = list(range(len(y)))
    ax.bar(x, y)
    ax.set_xticks(x)
    ax.set_xticklabels(keys)
    ax.set_ylabel('elapsed(sec)')


def draw_accuracy_elapsed(ax, data, limit=0.5):
    ax.set_title('Recall-elasped sec tradeoff')
    for library, values in data.items():
        values = sorted(filter(lambda x: x[0] > limit, values))
        ax.plot([acc for acc, _, _, _ in values], [elapsed for _, elapsed, _, _ in values], label=library)
    ax.set_xlabel('accuracy')
    ax.set_ylabel('elapsed(log scale)')
    ax.legend(loc='best')


def draw_index_size(ax, data):
    ax.set_title('index size(kb)')
    keys = data.keys()
    y = [min([kb for _, _, _, kb in data[k]]) for k in keys]
    x = list(range(len(y)))
    ax.bar(x, y)
    ax.set_xticks(x)
    ax.set_xticklabels(keys)
    ax.set_ylabel('kb')


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('result', help='result file path')
    args = parser.parse_args()

    if not os.path.exists(args.result):
        raise ValueError("Wrong result file path")

    visualize(parse(args.result), os.path.splitext(args.result)[0] + '.png')



# -*- coding: utf-8 -*-
import time
import logging
import logging.handlers

from buffalo.misc._log import PyBuffaloLog


NOTSET = 0
WARN = 1
INFO = 2
DEBUG = 3
TRACE = 4
__logger_holder = []

devnull = None


def set_log_level(lvl):
    for logger in __logger_holder:
        if lvl == NOTSET:
            logger.setLevel(logging.NOTSET)
        elif lvl == WARN:
            logger.setLevel(logging.WARN)
        elif lvl == INFO:
            logger.setLevel(logging.INFO)
        elif lvl == DEBUG:
            logger.setLevel(logging.DEBUG)
        else:
            logger.setLevel(logging.DEBUG)
    PyBuffaloLog().set_log_level(lvl)


def get_log_level():
    return PyBuffaloLog().get_log_level()


def get_logger(name=__file__, no_fileno=False):
    global __logger_holder
    logger = logging.getLogger(name)
    if logger.handlers:
        return logger

    lvl = get_log_level()
    if lvl == NOTSET:
        logger.setLevel(logging.NOTSET)
    elif lvl == WARN:
        logger.setLevel(logging.WARN)
    elif lvl == INFO:
        logger.setLevel(logging.INFO)
    elif lvl == DEBUG:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.DEBUG)

    sh = logging.StreamHandler()
    sh.setLevel(logging.DEBUG)
    if name == 'pbar':
        formatter = logging.Formatter('%(message)s')
    elif no_fileno:
        formatter = logging.Formatter('[%(levelname)-8s] %(asctime)s %(message)s', '%Y-%m-%d %H:%M:%S')
    else:
        formatter = logging.Formatter('[%(levelname)-8s] %(asctime)s [%(filename)s:%(lineno)d] %(message)s', '%Y-%m-%d %H:%M:%S')
    sh.setFormatter(formatter)

    logger.addHandler(sh)
    __logger_holder.append(logger)
    return logger


class ProgressBar(object):
    def __init__(self, level, **kwargs):
        self.lvl = level
        self._logger = get_logger('pbar', no_fileno=True)
        self.logger = self.get_logger_func(self._logger, self.lvl)
        for handle in self._logger.handlers:
            self.terminator = handle.terminator
            handle.terminator = ''
        self.initialize(**kwargs)

    def get_logger_func(self, logger, level):
        if level == INFO:
            return logger.info
        elif level == WARN:
            return logger.warn
        elif level == DEBUG:
            return logger.debug
        else:
            return logger.debug

    def initialize(self, **kwargs):
        if 'iterable' in kwargs:
            try:
                kwargs['total'] = len(kwargs['iterable'])
            except Exception:
                pass
            self.iterable = kwargs['iterable']
        if 'total' in kwargs:
            self.fmt = '\r[{desc}] {percent} {elapsed:0.1f}/{remains:0.1f}secs {ips:0,.2f}it/s'
        else:
            self.fmt = '\r[{desc}] {n_fmt:10,d}it {elapsed:0.1f}secs {ips:0,.2f}it/s'
        self.s_t = time.time()
        self.t = time.time()
        self.desc = kwargs.get('desc', 'PROGRESS')
        self.period_secs = kwargs.get('mininteral', 1)
        self.total = kwargs.get('total', -1)
        self.step = 0

    def get_msg(self):
        t = time.time()
        elapsed = t - self.s_t
        remains = 0
        percent = '0.00%'
        ips = 0.0
        if self.step > 0:
            remains = self.total / self.step * elapsed
            ips = self.step / elapsed
            percent = '%3.2f%s' % (elapsed / remains * 100, '%')
        msg = self.fmt.format(desc=self.desc,
                              percent=percent,
                              n_fmt=int(self.step),
                              total_fmt=self.total,
                              elapsed=elapsed,
                              remains=remains,
                              ips=ips)
        return msg

    def __enter__(self):
        self.logger(self.get_msg())
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.total != -1:
            self.step = self.total
        self.logger(self.get_msg() + '\n')

    def __iter__(self):
        delta = None
        cnts = 0
        period_iter = 0
        for obj in self.iterable:
            yield obj
            cnts += 1
            self.step += 1
            if period_iter != 0 and (cnts % period_iter == 0):
                self.logger(self.get_msg())
            else:
                delta = time.time() - self.s_t
                if delta > self.period_secs:
                    period_iter = cnts
        if self.total != -1:
            self.step = self.total
        self.logger(self.get_msg() + '\n')

    def update(self, step):
        self.step += step
        t = time.time()
        if self.t + self.period_secs < t:
            self.t = t
            self.logger(self.get_msg())

    def refresh(self):
        # backward-compatability
        pass


class supress_log_level(object):
    def __init__(self, log_level):
        self.desire_log_level = log_level
        self.original_log_level = get_log_level()

    def __enter__(self):
        set_log_level(self.desire_log_level)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        set_log_level(self.original_log_level)



import os
import sys
import gzip
import shutil
import tarfile
import argparse
import subprocess
from contextlib import closing

try:
    from urllib import urlretrieve
except ImportError:
    from urllib.request import urlretrieve  # Python 3


DATASETS = [
    'fashion-mnist-784-euclidean',
    'gist-960-euclidean',
    'glove-25-angular',
    'glove-50-angular',
    'glove-100-angular',
    'glove-200-angular',
    'mnist-784-euclidean',
    'random-xs-20-euclidean',
    'random-s-100-euclidean',
    'random-xs-20-angular',
    'random-s-100-angular',
    'sift-128-euclidean',
    'nytimes-256-angular',
    'youtube-40-angular',
    'youtube1m-40-angular',
]


def download(src, dst):
    if not os.path.exists(dst):
        urlretrieve(src, dst)


def get_dataset_fn(dataset):
    if not os.path.exists('data'):
        os.mkdir('data')
    return os.path.join('data', '%s.hdf5' % dataset)


def get_dataset(which, baseurl='http://ann-benchmarks.com/'):
    hdf5_fn = get_dataset_fn(which)
    try:
        url = '%s%s.hdf5' % (baseurl, which)
        download(url, hdf5_fn)
    except:
        raise IOError("Cannot download %s" % url)



if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--dataset', help='Which dataset',  choices=DATASETS)
    args = parser.parse_args()

    if args.dataset in ['youtube1m-40-angular', 'youtube-40-angular']:
        get_dataset(args.dataset, baseurl='https://arena.kakaocdn.net/n2/dataset/')
    else:
        get_dataset(args.dataset)



# cython: experimental_cpp_class_def=True, language_level=3
# distutils: language=c++
# -*- coding: utf-8 -*-


cdef extern from "buffalo/misc/log.hpp":
    cdef cppclass _BuffaloLogger "BuffaloLogger":
        void set_log_level(int)
        int get_log_level()


cdef class PyBuffaloLog:
    """CALS object holder"""
    cdef _BuffaloLogger* obj

    def __cinit__(self):
        self.obj = new _BuffaloLogger()

    def __dealloc__(self):
        del self.obj

    def set_log_level(self, lvl):
        self.obj.set_log_level(lvl)

    def get_log_level(self):
        return self.obj.get_log_level()



# -*- coding: utf-8 -*-
from buffalo.misc.aux import InputOptions, Option


class AlgoOption(InputOptions):
    def __init__(self, *args, **kwargs):
        super(AlgoOption, self).__init__(*args, **kwargs)

    def get_default_option(self):
        """Default options for Algo classes.

        :ivar bool evaluation_on_learning: Set True to do run evaluation on training phrase. (default: True)
        :ivar bool compute_loss_on_training: Set True to calculate loss on training phrase. (default: True)
        :ivar int early_stopping_rounds: The number of exceed epochs after reached minimum loss on training phrase. If set 0, it doesn't work. (default: 0)
        :ivar bool save_best: Whenver the loss improved, save the model.
        :ivar int evaluation_period: How often will do evaluation in epochs. (default: 1)
        :ivar int save_period: How often will do save_best routine in epochs. (default: 10)
        :ivar int random_seed: Random Seed
        :ivar dict validation: The validation options.
        """
        opt = {
            'evaluation_on_learning': True,
            'compute_loss_on_training': True,
            'early_stopping_rounds': 0,
            'save_best': False,
            'evaluation_period': 1,
            'save_period': 10,
            'random_seed': 0,
            'validation': {}
        }
        return opt

    def get_default_optimize_option(self):
        """Default options for optimize feature.

        :ivar str loss: Name of loss to monitor. (default: train_loss)
        """
        opt = {
            'loss': 'train_loss',
        }
        return opt

    def get_default_tensorboard_option(self):
        """Default options for tensorboard feature.

        :ivar str name: Name of graph name. (default: default)
        :ivar str root: Path where to make graph directory. (default: ./tb)
        :ivar str name_template: Name template for directory name. (default: {name}.{dtm})
        """
        opt = {
            'name': 'default',
            'root': './tb',
            'name_template': '{name}.{dtm}'
        }
        return opt

    def is_valid_option(self, opt):
        b = super().is_valid_option(opt)
        for f in ['num_workers']:
            if f not in opt:
                raise RuntimeError(f'{f} not defined')
        return b


class ALSOption(AlgoOption):
    def __init__(self, *args, **kwargs):
        super(ALSOption, self).__init__(*args, **kwargs)

    def get_default_option(self):
        """Options for Alternating Least Squares.

        :ivar bool adaptive_reg: Set True, for adaptive regularization. (default: False)
        :ivar bool save_factors: Set True, to save models. (default: False)
        :ivar bool accelerator: Set True, to accelerate training using GPU. (default: False)
        :ivar int d: The number of latent feature dimension. (default: 20)
        :ivar int num_iters: The number of iterations for training. (default: 10)
        :ivar int num_workers: The number of threads. (default: 1)
        :ivar int hyper_threads: The number of hyper threads when using cuda cores. (default: 256)
        :ivar float reg_u: The L2 regularization coefficient for user embedding matrix. (default: 0.1)
        :ivar float reg_i: The L2 regularization coefficient for item embedding matrix. (default: 0.1)
        :ivar float alpha: The coefficient of giving more weights to losses on positive samples. (default: 8)
        :ivar float eps: epsilon for numerical stability (default: 1e-10)
        :ivar float cg_tolerance: tolerance of conjugate gradient for early stopping iterations (default: 1e-10)
        :ivar str optimizer: The name of optimizer, should be in [llt, ldlt, manual_cg, eigen_cg, eigen_bicg, eigen_gmres, eigen_dgmres, eigen_minres]. (default: manual_cg)
        :ivar int num_cg_max_iters: The number of maximum iterations for conjuaget gradient optimizer. (default: 3)
        :ivar str model_path: Where to save model.
        :ivar dict data_opt: This options will be used to load data if given.
        """

        opt = super().get_default_option()
        opt.update({
            'adaptive_reg': False,
            'save_factors': False,
            'accelerator': False,

            'd': 20,
            'num_iters': 10,
            'num_workers': 1,
            'hyper_threads': 256,
            'num_cg_max_iters': 3,
            'reg_u': 0.1,
            'reg_i': 0.1,
            'alpha': 8,
            'optimizer': 'manual_cg',
            'cg_tolerance': 1e-10,
            'eps': 1e-10,

            'model_path': '',
            'data_opt': {}
        })
        return Option(opt)

    def get_default_optimize_option(self):
        """Optimization Options for ALS.

        :ivar str loss: Target loss to optimize.
        :ivar int max_trials: The maximum experiments for optimization. If not given, run forever.
        :ivar int min_trials: The minimum experiments before deploying model. (Since the best parameter may not be found after `min_trials`, the first best parameter is always deployed)
        :ivar bool deployment: Set True to train model with the best parameter. During the optimization, it try to dump the model which beated the previous best loss.
        :ivar bool start_with_default_parameters: If set to True, the loss value of the default parameter is used as the starting loss to beat.
        :ivar dict space: The parameter space definition. For more information, pleases reference hyperopt's express. Note) Due to hyperopt's `randint` does not provide lower value, we had to implement it a bait tricky. Pleases see optimize.py to check how we deal with `randint`.k
        """
        opt = super().get_default_optimize_option()
        opt.update({
            'loss': 'train_loss',
            'max_trials': 100,
            'min_trials': 0,
            'deployment': True,
            'start_with_default_parameters': True,
            'space': {
                'adaptive_reg': ['choice', ['adaptive_reg', [0, 1]]],
                'd': ['randint', ['d', 10, 30]],
                'reg_u': ['uniform', ['reg_u', 0.1, 1]],
                'reg_i': ['uniform', ['reg_i', 0.1, 1]],
                'alpha': ['randint', ['alpha', 1, 32]]
            }
        })
        return Option(opt)


class CFROption(AlgoOption):
    def __init__(self, *args, **kwargs):
        super(CFROption, self).__init__(*args, **kwargs)

    def get_default_option(self):
        """ Basic Options for CoFactor.

        :ivar int d: The number of latent feature dimension. (default: 20)
        :ivar int num_iters: The number of iterations for training. (default: 10)
        :ivar int num_workers: The number of threads. (default: 1)
        :ivar float reg_u: The L2 regularization coefficient for user embedding matrix. (default: 0.1)
        :ivar float reg_i: The L2 regularization coefficient for item embedding matrix. (default: 0.1)
        :ivar float reg_c: The L2 regularization coefficient for context embedding matrix. (default: 0.1)
        :ivar float eps: epsilon for numerical stability (default: 1e-10)
        :ivar float cg_tolerance: The tolerance for early stopping conjugate gradient optimizer. (default: 1e-10)
        :ivar float alpha: The coefficient of giving more weights to losses on positive samples. (default: 8.0)
        :ivar float l: The relative weight of loss on user-item relation over item-context relation. (default: 1.0)
        :ivar str optimizer: The name of optimizer, should be in [llt, ldlt, manual_cg, eigen_cg, eigen_bicg, eigen_gmres, eigen_dgmres, eigen_minres]. (default: manual_cg)
        :ivar int num_cg_max_iters: The number of maximum iterations for conjuaget gradient optimizer. (default: 3)
        :ivar str model_path: Where to save model. (default: '')
        :ivar dict data_opt: This options will be used to load data if given. (default: {})
        """
        opt = super().get_default_option()
        opt.update({
            'save_factors': False,
            'd': 20,
            'num_iters': 10,
            'num_workers': 1,
            'num_cg_max_iters': 3,

            'cg_tolerance': 1e-10,
            'eps': 1e-10,
            'reg_u': 0.1,
            'reg_i': 0.1,
            'reg_c': 0.1,
            'alpha': 8.0,
            'l': 1.0,

            'optimizer': 'manual_cg',
            'model_path': '',
            'data_opt': {}
        })
        return Option(opt)

    def get_default_optimize_option(self):
        """Optimization options for CoFactor.

        :ivar str loss: Target loss to optimize.
        :ivar int max_trials: Maximum experiments for optimization. If not given, run forever.
        :ivar int min_trials: Minimum experiments before deploying model. (Since the best parameter may not be found after `min_trials`, the first best parameter is always deployed)
        :ivar bool deployment(: Set True to train model with the best parameter. During the optimization, it try to dump the model which beated the previous best loss.
        :ivar bool start_with_default_parameters: If set to True, the loss value of the default parameter is used as the starting loss to beat.
        :ivar dict space: Parameter space definition. For more information, pleases reference hyperopt's express. Note) Due to hyperopt's `randint` does not provide lower value, we had to implement it a bait tricky. Pleases see optimize.py to check how we deal with `randint`.k
        """
        opt = super().get_default_optimize_option()
        opt.update({
            'loss': 'train_loss',
            'max_trials': 100,
            'min_trials': 0,
            'deployment': True,
            'start_with_default_parameters': True,
            'space': {
                'd': ['randint', ['d', 10, 30]],
                'reg_u': ['uniform', ['reg_u', 0.1, 1]],
                'reg_i': ['uniform', ['reg_i', 0.1, 1]],
                'reg_c': ['uniform', ['reg_i', 0.1, 1]],
                'alpha': ['randint', ['alpha', 1, 32]],
                'l': ['randint', ['alpha', 1, 32]]
            }
        })
        return Option(opt)

    def is_valid_option(self, opt):
        b = super().is_valid_option(opt)
        possible_optimizers = ["llt", "ldlt", "manual_cg", "eigen_cg", "eigen_bicg",
                               "eigen_gmres", "eigen_dgmres", "eigen_minres"]
        if opt.optimizer not in possible_optimizers:
            msg = f"optimizer ({opt.optimizer}) should be in {possible_optimizers}"
            raise RuntimeError(msg)
        return b


class BPRMFOption(AlgoOption):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def get_default_option(self):
        """Options for Bayesian Personalized Ranking Matrix Factorization.

        :ivar bool accelerator: Set True, to accelerate training using GPU. (default: False)
        :ivar bool use_bias: Set True, to use bias term for the model.
        :ivar int evaluation_period: (default: 100)
        :ivar int num_workers: The number of threads. (default: 1)
        :ivar int hyper_threads: The number of hyper threads when using cuda cores. (default: 256)
        :ivar int num_iters: The number of iterations for training. (default: 100)
        :ivar int d: The number of latent feature dimension. (default: 20)
        :ivar bool update_i: Set True, to update positive item feature. (default: True)
        :ivar bool update_j: Set True, to update negative item feature. (default: True)
        :ivar float reg_u: The L2 regularization coefficient for user embedding matrix. (default: 0.025)
        :ivar float reg_i: The L2 regularization coefficient for positive item embedding matrix. (default: 0.025)
        :ivar float reg_j: The L2 regularization coefficient for negative item embedding matrix. (default: 0.025)
        :ivar float reg_b: The L2 regularization coefficient for bias term. (default: 0.025)
        :ivar str optimizer: The name of optimizer, should be one of [sgd, adam]. (default: sgd)
        :ivar float lr: The learning rate.
        :ivar float min_lr: The minimum of learning rate, to prevent going to zero by learning rate decaying. (default: 0.0001)
        :ivar float beta1: The parameter of Adam optimizer. (default: 0.9)
        :ivar float beta2: The parameter of Adam optimizer. (default: 0.999)
        :ivar bool per_coordinate_normalize: This is a bit tricky option for Adam optimizer. Before update factors with graidents, do normalize gradients per class by its number of contributed samples. (default: False)
        :ivar int num_negative_samples: The number of negaitve samples. (default: 1)
        :ivar float sampling_power: This paramemter control sampling distribution. When it set to 0, it draw negative items from uniform distribution, while to set 1, it draw from the given data popularation. (default: 0.0)
        :ivar bool random_positive: Set True, to draw positive sample uniformly instead of using straight forward positive sample, only implemented in cuda mode, according to the original paper, set True, but we found out False usually produces better results) (default: False)
        :ivar bool verify_neg: Set True, to ensure negative sample does not belong to positive samples. (default True)
        :ivar str model_path: Where to save model.
        :ivar dict data_opt: This options will be used to load data if given.
        """
        opt = super().get_default_option()
        opt.update({
            'accelerator': False,
            'use_bias': True,
            'evaluation_period': 100,
            'num_workers': 1,
            'hyper_threads': 256,
            'num_iters': 100,
            'd': 20,
            'update_i': True,
            'update_j': True,
            'reg_u': 0.025,
            'reg_i': 0.025,
            'reg_j': 0.025,
            'reg_b': 1.0,

            'optimizer': 'sgd',
            'lr': 0.002,
            'min_lr': 0.0001,
            'beta1': 0.9,
            'beta2': 0.999,
            'eps': 1e-10,

            'per_coordinate_normalize': False,
            'num_negative_samples': 1,
            'sampling_power': 0.0,
            'verify_neg': True,
            'random_positive': False,

            'model_path': '',
            'data_opt': {}
        })
        return Option(opt)

    def get_default_optimize_option(self):
        """Optimization options for BPRMF.
        """
        opt = super().get_default_optimize_option()
        opt.update({
            'loss': 'train_loss',
            'max_trials': 100,
            'min_trials': 0,
            'deployment': True,
            'start_with_default_parameters': True,
            'space': {
                'd': ['randint', ['d', 10, 30]],
                'reg_u': ['uniform', ['reg_u', 0.1, 1]],
                'reg_i': ['uniform', ['reg_i', 0.1, 1]]
            }
        })
        return Option(opt)


class W2VOption(AlgoOption):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def get_default_option(self):
        """Options for Word2Vec.

        :ivar bool evaluation_on_learning: Set True to do run evaluation on training phrase. (default: False)
        :ivar int num_workers: The number of threads. (default: 1)
        :ivar int num_iters: The number of iterations for training. (default: 100)
        :ivar int d: The number of latent feature dimension. (default: 20)
        :ivar int window: The window size. (default: 5)
        :ivar int min_count: The minimum required frequency of the words to use training vocabulary. (default: 5)
        :ivar float sample: The sampling ratio to downsample the frequent words. (default: 0.001)
        :ivar float lr: The learning rate.
        :ivar int num_negative_samples: The number of negaitve samples. (default: 5)
        :ivar str model_path: Where to save model.
        :ivar dict data_opt: This options will be used to load data if given.
        """
        opt = super().get_default_option()
        opt.update({
            'evaluation_on_learning': False,

            'num_workers': 1,
            'num_iters': 3,
            'd': 20,
            'window': 5,
            'min_count': 5,
            'sample': 0.001,

            'lr': 0.025,
            'min_lr': 0.0001,

            'num_negative_samples': 5,

            'model_path': '',
            'data_opt': {}
        })
        return Option(opt)

    def get_default_optimize_option(self):
        """Optimization options for W2V
        """
        #
        opt = super().get_default_optimize_option()
        opt.update({
            'loss': 'train_loss',
            'max_trials': 100,
            'min_trials': 0,
            'deployment': True,
            'start_with_default_parameters': True,
            'space': {
                'd': ['randint', ['d', 10, 30]],
                'window': ['randint', ['window', 2, 8]],
                'num_negative_samples': ['randint', ['alpha', 1, 12]]
            }
        })
        return Option(opt)



# -*- coding: utf-8 -*-
import numpy as np
from buffalo.parallel._core import quickselect


class Evaluable(object):
    def __init__(self, *args, **kargs):
        pass

    def prepare_evaluation(self):
        if not self.opt.validation or not self.data.has_group('vali'):
            return
        if hasattr(self.data, 'vali_data') is False:
            self.data._prepare_validation_data()

    def show_validation_results(self):
        results = self.get_validation_results()
        if not results:
            return 'No validation results'
        return 'Validation results: ' + ', '.join(f'{k}: {v:0.5f}' for k, v in results.items())

    def get_validation_results(self):
        if not self.opt.validation or not self.data.has_group('vali'):
            return

        results = {}
        results.update(self._evaluate_ranking_metrics())
        results.update(self._evaluate_score_metrics())
        return results

    def get_topk(self, scores, k, sorted=True, num_threads=4):
        # NOTE: Is it necessary condition?
        # assert k < scores.shape[1], f"k ({k}) should be smaller than cols ({scores.shape[1]})"
        is_many = True
        if len(scores.shape) == 1:
            scores = scores.reshape(1, scores.shape[0])
            is_many = False
        k = min(k, scores.shape[1])
        assert k > 0, f'k({k}) or cols({scores.shape[1]}) should be greater than 0'
        result = np.empty(shape=(scores.shape[0], k), dtype=np.int32)
        quickselect(scores, result, sorted, num_threads)
        return result if is_many else result[0]

    def _evaluate_ranking_metrics(self):
        if hasattr(self.data, 'vali_data') is False:
            self.prepare_evaluation()

        batch_size = self.opt.validation.get('batch', 128)
        topk = self.opt.validation.topk

        gt = self.data.vali_data['vali_gt']
        rows = self.data.vali_data['vali_rows']
        validation_seen = self.data.vali_data['validation_seen']
        validation_max_seen_size = self.data.vali_data['validation_max_seen_size']
        num_items = self.data.get_header()['num_items']

        # can significantly save evaluation time
        if self.opt.validation.eval_samples:
            size = min(self.opt.validation.eval_samples, len(rows))
            rows = np.random.choice(rows, size=size, replace=False)

        NDCG = 0.0
        AP = 0.0
        HIT = 0.0
        AUC = 0.0
        N = 0.0

        idcgs = np.cumsum(1.0 / np.log2(np.arange(2, topk + 2)))
        dcgs = 1.0 / np.log2(np.arange(2, topk + 2))

        def filter_seen_items(_topk, seen, topk):
            ret = []
            for t in _topk:
                if t not in seen:
                    ret.append(t)
                    if len(ret) >= topk:
                        break
            return ret

        for index in range(0, len(rows), batch_size):
            recs = self._get_topk_recommendation(rows[index:index + batch_size],
                                                 topk=topk + validation_max_seen_size)
            for row, _topk in recs:
                seen = validation_seen.get(row, set())

                if len(seen) == 0:
                    continue

                _topk = filter_seen_items(_topk, seen, topk)
                _gt = gt[row]

                # accuracy
                hit = len(set(_topk) & _gt) / len(_gt)
                HIT += hit

                # ndcg, map
                idcg = idcgs[min(len(_gt), len(_topk)) - 1]
                dcg = 0.0
                hit, miss, ap = 0.0, 0.0, 0.0

                # AUC
                num_pos_items = len(_gt)
                num_neg_items = num_items - num_pos_items
                auc = 0.0

                for i, r in enumerate(_topk):
                    if r in _gt:
                        hit += 1
                        ap += (hit / (i + 1.0))
                        dcg += dcgs[i]
                    else:
                        miss += 1
                        auc += hit
                auc += ((hit + num_pos_items) / 2.0) * (num_neg_items - miss)
                auc /= (num_pos_items * num_neg_items)

                ndcg = dcg / idcg
                NDCG += ndcg
                ap /= min(len(_gt), len(_topk))
                AP += ap
                N += 1.0
                AUC += auc
        NDCG /= N
        AP /= N
        ACC = HIT / N
        AUC = AUC / N
        ret = {'ndcg': NDCG, 'map': AP, 'accuracy': ACC, 'auc': AUC}
        return ret

    def _evaluate_score_metrics(self):
        if hasattr(self.data, 'vali_data') is False:
            self.prepare_evaluation()

        vali_data = self.data.vali_data
        row = vali_data['row']
        col = vali_data['col']
        val = vali_data['val']
        scores = self._get_scores(row, col)
        ERROR = 0.0
        RMSE = 0.0
        for r, c, p, v in zip(row, col, scores, val):
            err = p - v
            ERROR += abs(err)
            RMSE += err * err
        RMSE /= len(scores)
        RMSE = RMSE ** 0.5
        ERROR /= len(scores)
        return {'rmse': RMSE, 'error': ERROR}



# cython: experimental_cpp_class_def=True, language_level=3
# distutils: language=c++
# -*- coding: utf-8 -*-
import cython
from libcpp cimport bool
from libcpp.pair cimport pair
from libcpp.string cimport string
from libc.stdint cimport int32_t, int64_t

import numpy as np
cimport numpy as np


cdef extern from "buffalo/cuda/bpr/bpr.hpp" namespace "cuda_bpr":
    cdef cppclass CuBPR:
        CuBPR() nogil except +
        bool init(string) nogil except +
        void set_placeholder(int64_t* indptr, size_t batch_size) nogil except +
        void set_cumulative_table(int64_t*) nogil except +
        void initialize_model(float*, int,
                              float*, float*, int, int64_t, bool) nogil except +
        pair[double, double] partial_update(int, int,
                                            int64_t*, int32_t*) nogil except +
        double compute_loss(int, int32_t*, int32_t*, int32_t*) nogil except +
        int get_vdim() nogil except +
        void synchronize(bool)


cdef class CyBPR:
    """CuBPR object holder"""
    cdef CuBPR* obj  # Cuda-ALS object

    def __cinit__(self):
        self.obj = new CuBPR()

    def __dealloc__(self):
        del self.obj

    def init(self, opt_path):
        return self.obj.init(opt_path)

    def initialize_model(self,
                         np.ndarray[np.float32_t, ndim=2] P,
                         np.ndarray[np.float32_t, ndim=2] Q,
                         np.ndarray[np.float32_t, ndim=2] Qb,
                         num_nnz, set_gpu=False):
        self.obj.initialize_model(&P[0, 0], P.shape[0],
                                  &Q[0, 0], &Qb[0, 0], Q.shape[0], num_nnz, set_gpu)

    def set_placeholder(self, np.ndarray[np.int64_t, ndim=1] indptr, batch_size):
        self.obj.set_placeholder(&indptr[0], batch_size)

    def set_cumulative_table(self, np.ndarray[np.int64_t, ndim=1] sampling_table, size):
        self.obj.set_cumulative_table(&sampling_table[0])

    def get_vdim(self):
        return self.obj.get_vdim()

    def synchronize(self, device_to_host):
        self.obj.synchronize(device_to_host)

    def update_parameters(self):
        self.synchronize(True)

    def wait_until_done(self):
        return

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def add_jobs(self,
                 int start_x,
                 int next_x,
                 np.ndarray[np.int64_t, ndim=1] indptr,
                 np.ndarray[np.int32_t, ndim=1] keys):
        return self.obj.partial_update(start_x,
                                       next_x,
                                       &indptr[0],
                                       &keys[0])

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def compute_loss(self,
                     np.ndarray[np.int32_t, ndim=1] user,
                     np.ndarray[np.int32_t, ndim=1] pos,
                     np.ndarray[np.int32_t, ndim=1] neg):
        return self.obj.compute_loss(user.shape[0], &user[0], &pos[0], &neg[0])



import os
import sys
import abc
import time
import random
import psutil
import logging
import argparse
import resource
import multiprocessing

import h5py
import numpy
import nmslib

from download_dataset import get_dataset_fn, DATASETS
from n2 import HnswIndex

try:
    xrange
except NameError:
    xrange = range


CACHE_DIR = './cache'
RESULT_DIR = './result'
GT_SIZE = 100


logging.basicConfig(format='%(message)s')
n2_logger = logging.getLogger("n2_benchmark")
n2_logger.setLevel(logging.INFO)


# Set resource limits to prevent memory bombs
memory_limit = 64 * 2**30
soft, hard = resource.getrlimit(resource.RLIMIT_DATA)
if soft == resource.RLIM_INFINITY or soft >= memory_limit:
    n2_logger.debug('resetting memory limit from {0} to {1}. '.format(soft, memory_limit))
    resource.setrlimit(resource.RLIMIT_DATA, (memory_limit, hard))


class BaseANN(object):
    @abc.abstractmethod
    def fit(self, X):
        pass

    @abc.abstractmethod
    def query(self, v, n):
        pass

    @abc.abstractmethod
    def __str__(self):
        pass

    def get_memory_usage(self):
        return psutil.Process().memory_info().rss / 1024


class N2(BaseANN):
    def __init__(self, m, ef_construction, n_threads, ef_search, metric):
        self.name = "N2_M%d_efCon%d_n_thread%s_efSearch%d" % (m, ef_construction, n_threads, ef_search)
        self._m = m
        self._m0 = m * 2
        self._ef_construction = ef_construction
        self._n_threads = n_threads
        self._ef_search = ef_search
        self._index_name = os.path.join(CACHE_DIR, "index_n2_%s_M%d_efCon%d_n_thread%s_datasz%d"
                                        % (args.dataset, m, ef_construction, n_threads, args.data_size))
        self._metric = metric

    def fit(self, X):
        if self._metric == 'euclidean':
            self._n2 = HnswIndex(X.shape[1], 'L2')
        else:
            self._n2 = HnswIndex(X.shape[1])

        if os.path.exists(self._index_name):
            n2_logger.info("Loading index from file")
            self._n2.load(self._index_name, use_mmap=False)
            return

        n2_logger.debug("Create Index")
        for i, x in enumerate(X):
            self._n2.add_data(x)
        self._n2.build(m=self._m, max_m0=self._m0, ef_construction=self._ef_construction, n_threads=self._n_threads)
        self._n2.save(self._index_name)

    def query(self, v, n):
        return self._n2.search_by_vector(v, n, self._ef_search)

    def __str__(self):
        return self.name


class NmslibHNSW(BaseANN):
    def __init__(self, m, ef_construction, n_threads, ef_search, metric):
        self.name = "nmslib_M%d_efCon%d_n_thread%s_efSearch%d" % (m, ef_construction, n_threads, ef_search)
        self._index_param = [
            'M=%d' % m,
            'indexThreadQty=%d' % n_threads,
            'efConstruction=%d' % ef_construction,
            'post=0', 'delaunay_type=2']
        self._query_param = ['efSearch=%d' % ef_search]
        self._index_name = os.path.join(CACHE_DIR, "index_nmslib_%s_M%d_efCon%d_n_thread%s_datasz%d"
                                        % (args.dataset, m, ef_construction, n_threads, args.data_size))
        self._metric = {'angular': 'cosinesimil', 'euclidean': 'l2'}[metric]

    def fit(self, X):
        self._index = nmslib.init(self._metric, [], "hnsw", nmslib.DataType.DENSE_VECTOR, nmslib.DistType.FLOAT)

        if os.path.exists(self._index_name):
            logging.debug("Loading index from file")
            nmslib.loadIndex(self._index, self._index_name)
        else:
            logging.debug("Create Index")
            for i, x in enumerate(X):
                self._index.addDataPoint(i, x)

            nmslib.createIndex(self._index, self._index_param)
            nmslib.saveIndex(self._index, self._index_name)

        nmslib.setQueryTimeParams(self._index, self._query_param)

    def query(self, v, n):
        return nmslib.knnQuery(self._index, n, v)

    def free_index(self):
        nmslib.freeIndex(self._index)

    def __str__(self):
        return self.name


def run_algo(args, library, algo, results_fn):
    pool = multiprocessing.Pool()
    X_train, X_test, corrects = load_dataset(args.dataset)
    pool.close()
    pool.join()

    memory_usage_before = algo.get_memory_usage()
    t0 = time.time()
    algo.fit(X_train)
    build_time = time.time() - t0
    index_size_kb = algo.get_memory_usage() - memory_usage_before
    n2_logger.info('Built index in {0}, Index size: {1}KB'.format(build_time, index_size_kb))

    best_search_time = float('inf')
    best_precision = 0.0  # should be deterministic but paranoid
    try_count = args.try_count
    for i in xrange(try_count):  # Do multiple times to warm up page cache, use fastest
        results = []
        search_time = 0.0
        for j, v in enumerate(X_test):
            sys.stderr.write("[%d/%d][algo: %s] Querying: %d / %d \r"
                             % (i+1, try_count, str(algo), j+1, len(X_test)))
            t0 = time.time()
            found = algo.query(v, GT_SIZE)
            search_time += (time.time() - t0)

            results.append(len(set(found).intersection(corrects[j])))

            if len(found) < len(corrects[j]):
                n2_logger.debug('found: {0}, correct: {1}'.format(len(found), len(corrects[j])))

        sys.stderr.write("\n")

        k = float(sum(results))
        search_time /= len(X_test)
        precision = k / (len(X_test) * GT_SIZE)
        best_search_time = min(best_search_time, search_time)
        best_precision = max(best_precision, precision)
        n2_logger.debug('[%d/%d][algo: %s] search time: %s, precision: %.5f'
                        % (i+1, try_count, str(algo), str(search_time), precision))

    output = '\t'.join(map(str, [library, algo.name, build_time, best_search_time, best_precision, index_size_kb]))
    with open(results_fn, 'a') as f:
        f.write(output + '\n')

    n2_logger.info('Summary: {0}'.format(output))


def load_dataset(which):
    hdf5_fn = get_dataset_fn(which)
    f = h5py.File(hdf5_fn, 'r')
    X_train = numpy.array(f['train'])
    X_test = numpy.array(f['test'])
    corrects = f['neighbors']
    return X_train, X_test, corrects


def get_fn(file_type, args, base=CACHE_DIR):
    fn = '%s_%s_%d_%d_%d' % (os.path.join(base, file_type), args.dataset,
                             args.data_size, args.test_size, args.random_state)
    return fn


def run(args):
    results_fn = get_fn('result', args, base=RESULT_DIR) + '.txt'

    index_params = [(12, 100)]
    query_params = [25, 50, 100, 250, 500, 750, 1000, 1500, 2500, 5000, 10000]

    algos = {
        'n2': [N2(M, ef_con, args.n_threads, ef_search, args.distance)
               for M, ef_con in index_params
               for ef_search in query_params],
        'nmslib': [NmslibHNSW(M, ef_con, args.n_threads, ef_search, args.distance)
                   for M, ef_con in index_params
                   for ef_search in query_params],
    }

    if args.algo:
        algos = {args.algo: algos[args.algo]}

    algos_flat = [(k, v) for k, vals in algos.items() for v in vals]
    random.shuffle(algos_flat)
    n2_logger.debug('order: %s' % str([a.name for l, a in algos_flat]))

    for library, algo in algos_flat:
        # Spawn a subprocess to force the memory to be reclaimed at the end
        p = multiprocessing.Process(target=run_algo, args=(args, library, algo, results_fn))
        p.start()
        p.join()


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--distance', help='Distance metric', default='angular', choices=['angular', 'euclidean'])
    parser.add_argument('--try_count', help='Number of test attempts', type=int, default=3)
    parser.add_argument('--dataset', help='Which dataset',  default='glove', choices=DATASETS)
    parser.add_argument('--data_size', help='Maximum # of data points (0: unlimited)', type=int, default=0)
    parser.add_argument('--test_size', help='Maximum # of data queries', type=int, default=10000)
    parser.add_argument('--n_threads', help='Number of threads', type=int, default=10)
    parser.add_argument('--random_state', help='Random seed', type=int, default=3)
    parser.add_argument('--algo', help='Algorithm', type=str, choices=['n2', 'nmslib'])
    parser.add_argument('--verbose', '--v', help='print verbose log', type=bool, default=False)
    args = parser.parse_args()

    if not os.path.exists(get_dataset_fn(args.dataset)):
        raise IOError('Please download the dataset')

    if not os.path.exists(CACHE_DIR):
        os.makedirs(CACHE_DIR)

    if not os.path.exists(RESULT_DIR):
        os.makedirs(RESULT_DIR)

    if args.verbose:
        n2_logger.setLevel(logging.DEBUG)

    numpy.random.seed(args.random_state)

    run(args)
